package rwth.i2.ltlrv.afastate.base;

import java.util.HashSet;
import java.util.Set;

import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.base.AbstractFormula;

/**
 * Abstract superclass of all formulae. For ease of use only.
 * 
 * @author Eric Bodden
 */
public abstract class AbstractAFAState extends AbstractFormula implements IAFAState {
    
    /**
     * {@inheritDoc}
     */
    public void transition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
        //update the context according to the current layer
    	context = updateContext(context);
        //calculate the transition based on this new context
    	doTransition(context,currentProps,currentBinding,result);
    }

    /**
     * Updates the context according to the current layer.
     * This means it adds all variables provided by the current layer to the context.
     * @param context the context of the last (earlier) temporal layer 
     * @return <code>context</code> <i>join</i> <code>provides(this)</code> 
     * @see IAFAState#provides(Set)
     */
    protected Set<String> updateContext(Set<String> context) {
        Set<String> localContext = new HashSet<String>();
        localContext.addAll(context);
        ((IAFAState)this).provides(localContext);
        return localContext;
    }

    /**
     * Performs the actual transition based on the current context.
     * @param context @see #transition(Set, PropositionSet, WeakValuesMap, Set)
     * @param currentProps @see #transition(Set, PropositionSet, WeakValuesMap, Set)
     * @param currentBinding @see #transition(Set, PropositionSet, WeakValuesMap, Set)
     * @param result @see #transition(Set, PropositionSet, WeakValuesMap, Set)
     * @see IAFAState#transition(Set, PropositionSet, WeakValuesMap, Set)
     * @see #transition(Set, PropositionSet, WeakValuesMap, Set)
     * @see #updateContext(Set)
     */
    protected abstract void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result);

    /**
     * {@inheritDoc}
     */
    public void provides(Set<String> context) {
        providesPos(context);
        providesNeg(context);
    }

    /**
     * Implementation of {@link IAFAState#validate()}. This creates an empty context and then
     * recurses further using {@link #validate(Set)}.
     * @throws ValidationException @inheritDoc
     */
    public void validate() throws ValidationException {
        Set<String> context = new HashSet<String>();
        validate(context);
    }
    
    /**
     * To be implemented by actual formulae as described in {@link IAFAState#validate()}.
     * @param context Here the context is passed in from the outer (earlier) temporal layers. 
     * @throws ValidationException @see #validate()
     */
    public abstract void validate(Set<String> context) throws ValidationException;
    
}
