package rwth.i2.ltlrv.afastate.base;

import java.util.Set;

import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IBinaryAFAState;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;

/**
 * BinaryAFAState - Abstract superclass of all binary formulae.
 * 
 * @author Eric Bodden
 */
public abstract class BinaryAFAState extends AbstractAFAState implements IBinaryAFAState {

    /** the subformulae */
    protected IFormula subformula1, subformula2;
    
    protected IAFAState substate1, substate2;
    
    /**
     * @param subformula1 left subformula
     * @param subformula2 right subformula
     */
    public BinaryAFAState(IFormula subformula1, IFormula subformula2) {
        this.subformula1 = subformula1;
        this.subformula2 = subformula2;
    }

    /**
     * {@inheritDoc}
     */
    public IFormula getSubformula1() {
        return subformula1;
    }

    /**
     * {@inheritDoc}
     */
    public IFormula getSubformula2() {
        return subformula2;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    //simply resurse
    public boolean equals(Object oth) {
        if (this == oth) {
            return true;
        }
    
        if (oth == null) {
            return false;
        }
    
        if (oth.getClass() != getClass()) {
            return false;
        }
    
        BinaryAFAState other = (BinaryAFAState) oth;
        if (this.subformula1 == null) {
            if (other.subformula1 != null) {
                return false;
            }
        } else {
            if (!this.subformula1.equals(other.subformula1)) {
                return false;
            }
        }
        if (this.subformula2 == null) {
            if (other.subformula2 != null) {
                return false;
            }
        } else {
            if (!this.subformula2.equals(other.subformula2)) {
                return false;
            }
        }
    
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    //simply resurse
    public int hashCode() {
        final int PRIME = 1000003;
        int result = 0;
        if (subformula1 != null) {
            result = PRIME * result + subformula1.hashCode();
        }
        if (subformula2 != null) {
            result = PRIME * result + subformula2.hashCode();
        }
    
        return result;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    //simply resurse
    public String toString() {
        return "(" +
            this.subformula1.toString() +
            ") " + 
            this.symbol() +
            " (" +             
            this.subformula2.toString() +
            ")";
    }

    /**
     * Default implementation.
     * @param variableKind TODO
     * @see IAFAState#unboundVariablesAtCurrentJoinpoint(VariableKind, Set)
     */
    //TODO still needed?
    public void unboundVariablesAtCurrentJoinpoint(VariableKind variableKind, Set<String> result) {
        //only defined for AFA terms
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //simply recurse
        ((IAFAState)this.subformula1).unboundVariablesAtCurrentJoinpoint(variableKind, result);
        ((IAFAState)this.subformula2).unboundVariablesAtCurrentJoinpoint(variableKind, result);
    }

    
    /**
     * {@inheritDoc}
     */
    public void requires(Set<String> result) {
        //only defined for AFA terms
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //simply recurse and join
        ((IAFAState)this.subformula1).requires(result);
        ((IAFAState)this.subformula2).requires(result);
    }
 

    /**
     * {@inheritDoc}
     */
    @Override
    public void validate(Set<String> context) throws ValidationException {
        //only defined for AFA terms
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        assert (this.subformula1 instanceof AbstractAFAState) && (this.subformula2 instanceof AbstractAFAState);
        //build local context
        Set<String> localContext = updateContext(context);
        //validate
//    	Set<String> req = new HashSet<String>();
//        ((IAFAState)this.subformula1).requires(req);
//        ((IAFAState)this.subformula2).requires(req);
//        //TODO validation on Proposition level might suffice
//        for (String varName : req) {
//			if(!localContext.contains(varName)) {
//				 throw new ValidationException(varName,this);
//			}
//		}
        //recurse
        ((AbstractAFAState)this.subformula1).validate(localContext);
        ((AbstractAFAState)this.subformula2).validate(localContext);
    }
    
    /**
     * {@inheritDoc}
     */
    //by default, simply recurse
    public IAFAState negationNormalForm() {
        return reconstruct(subformula1.negationNormalForm(),subformula2.negationNormalForm());
    }
    
    /**
     * {@inheritDoc}
     */
    public IAFAState specializeBindings(WeakValuesMap<String, Object> bindings) {
        //only defined for AFA terms
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState);
        //simply recurse
        return reconstruct(
                ((IAFAState)subformula1).specializeBindings(bindings),
                ((IAFAState)subformula2).specializeBindings(bindings)        
        );
    }
}
