package rwth.i2.ltlrv.afastate.base;

import java.util.Set;

import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IUnaryAFAState;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;

/**
 * UnaryAFAState - Abstract superclass of all unary formulae.
 * 
 * @author Eric Bodden
 */
public abstract class UnaryAFAState extends AbstractAFAState implements IUnaryAFAState {

    /** the subformula  */
    protected IFormula subformula; 
    
    /**
     * @param subformula the subformula
     */
    public UnaryAFAState(IFormula subformula) {
        this.subformula = subformula;
    }

    /**
     * {@inheritDoc}
     */
    public IFormula getSubformula() {
        return subformula;
    }
    
    /**
     * {@inheritDoc}
     */
    public void requires(Set<String> result) {
        //only defined for AFA terms
        assert (this.subformula instanceof IAFAState);
        //simply recurse 
        ((IAFAState)this.subformula).requires(result);
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public void validate(Set<String> context) throws ValidationException {
        //only defined for AFA terms
        assert (this.subformula instanceof IAFAState);
        assert (this.subformula instanceof AbstractAFAState);
//        //TODO is it true, that we don't need to update the context yet check anything here?
        Set<String> localContext = updateContext(context);
//        Set<String> req = new HashSet<String>();
//        ((IAFAState)this.subformula).requires(req);
//        //TODO validation on Proposition level might suffice
//        for (String varName : req) {
//            if(!localContext.contains(varName)) {
//                 throw new ValidationException(varName,(IAFAState)this.subformula);
//            }
//        }
        //recurse
        ((AbstractAFAState)this.subformula).validate(localContext);
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    //simply recurse
    public boolean equals(Object oth) {
        if (this == oth) {
            return true;
        }
    
        if (oth == null) {
            return false;
        }
    
        if (oth.getClass() != getClass()) {
            return false;
        }
    
        UnaryAFAState other = (UnaryAFAState) oth;
        if (this.subformula == null) {
            if (other.subformula != null) {
                return false;
            }
        } else {
            if (!this.subformula.equals(other.subformula)) {
                return false;
            }
        }
    
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    //simply recurse
    public int hashCode() {
        final int PRIME = 1000003;
        int result = 0;
        if (subformula != null) {
            result = PRIME * result + subformula.hashCode();
        }
    
        return result;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    //simply recurse
    public String toString() {
        return this.symbol() +
            "(" +
            this.subformula.toString() +
            ")";
    }

    /**
     * Default implementation.
     * @param variableKind TODO
     * @see IAFAState#unboundVariablesAtCurrentJoinpoint(VariableKind, Set)
     */
    //TODO still needed?
    public void unboundVariablesAtCurrentJoinpoint(VariableKind variableKind, Set<String> result) {
        //only defined for AFA terms
        assert this.subformula instanceof IAFAState;
        //simply recurse
        ((IAFAState)this.subformula).unboundVariablesAtCurrentJoinpoint(variableKind, result);
    }
    
    /**
     * {@inheritDoc}
     */
    //by default, simply recurse
    public IAFAState negationNormalForm() {
        return reconstruct(subformula.negationNormalForm());
    }
    
    /**
     * {@inheritDoc}
     */
    public IAFAState specializeBindings(WeakValuesMap<String, Object> bindings) {
        //only defined for AFA terms
        assert this.subformula instanceof IAFAState;
        //simply recurse
        return reconstruct(((IAFAState)subformula).specializeBindings(bindings));
    }
    
}
