package rwth.i2.ltlrv.afastate.impl;

import java.util.HashSet;
import java.util.Set;

import rwth.i2.ltlrv.afastate.base.UnaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAnd;
import rwth.i2.ltlrv.afastate.interfaze.IFF;
import rwth.i2.ltlrv.afastate.interfaze.INext;
import rwth.i2.ltlrv.afastate.interfaze.INot;
import rwth.i2.ltlrv.afastate.interfaze.IOr;
import rwth.i2.ltlrv.afastate.interfaze.IProposition;
import rwth.i2.ltlrv.afastate.interfaze.IRelease;
import rwth.i2.ltlrv.afastate.interfaze.ITT;
import rwth.i2.ltlrv.afastate.interfaze.IUntil;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;
import rwth.i2.ltlrv.util.SetUtils;

/**
 * Not - Implements the <i>Not</i> operator.
 * This holds the major implementation of bringing a formula into
 * negation normal form.
 * 
 * @author Eric Bodden
 * @see rwth.i2.ltlrv.formula.interfaze.IFormula#negationNormalForm()
 */
public class Not extends UnaryAFAState implements INot {

    /**
     * A negated form of the subformula passed in.
     * @param subFormula the subformula
     */
    public Not(IFormula subFormula) {
        super(subFormula);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public IAFAState negationNormalForm() {
        //This is the main implementation of the generation of a negation normal form.
        
        if(subformula instanceof ITT) {
            //!TT = FF
            return factory.FF();
        } else if(subformula instanceof IFF) {
            //!FF = TT
            return factory.TT();
        } else if(subformula instanceof INext) {
            //!X(sub) = X(!sub)
            return factory.X(factory.Not(((INext)subformula).getSubformula()).negationNormalForm());
        } else if(subformula instanceof INot) {
            //!!sub = sub
            return ((INot)subformula).getSubformula().negationNormalForm();
        } else if(subformula instanceof IUntil) {
            //!(s1 U s2) = !s1 R !s2
            return factory.Release(
                    factory.Not(((IUntil)subformula).getSubformula1()),
                    factory.Not(((IUntil)subformula).getSubformula2())
            ).negationNormalForm();
        } else if(subformula instanceof IRelease) {
            //!(s1 R s2) = !s1 U !s2
            return factory.Until(
                    factory.Not(((IRelease)subformula).getSubformula1()),
                    factory.Not(((IRelease)subformula).getSubformula2())
            ).negationNormalForm();
        } else if(subformula instanceof IAnd) {
            //!(a && b) = !a || !b
            return factory.Or(
                    factory.Not(((IAnd)subformula).getSubformula1()),
                    factory.Not(((IAnd)subformula).getSubformula2())
            ).negationNormalForm();
        } else if(subformula instanceof IOr) {
            //!(a || b) = !a && !b
            return factory.And(
                    factory.Not(((IOr)subformula).getSubformula1()),
                    factory.Not(((IOr)subformula).getSubformula2())
            ).negationNormalForm();
        } else if(subformula instanceof IProposition) {
            //!prop = !prop
            return this;
        } else {
            //simply recurse for other cases
            //TODO get rid of instanceof
            return super.negationNormalForm();
        }
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
        //only subformulae implementing IAFAState allowed at this stage
        assert this.subformula instanceof IAFAState;
        //this state is final if its subformula is nonfinal
        return !((IAFAState)this.subformula).isFinalStateInAFA();
    }

	/**
	 * {@inheritDoc}
	 */
	@Override
    public void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
		assert subformula instanceof IProposition;        
        //exploit NNF: the subformula can only be a proposition; thus its transition will be either the set representing TT
        //({{}}) or FF ({})
		Set<Set<IAFAState>> tempResult = new HashSet<Set<IAFAState>>();
		((IAFAState)subformula).transition(context,currentProps,currentBinding, tempResult);
		if(tempResult.size()>0) {
            //if TT (={{}})
			//if true, the result is the empty set
		} else {
            //if FF (={})
			//if false, add an empty set, which represents TT
			result.add(SetUtils.clauseTT());
		}		
	}
	
    /**
     * {@inheritDoc}
     */
    @SuppressWarnings("unused")
	public void providesPos(Set<String> result) {
		//negated propositions provide nothing positively
	}

    /**
     * {@inheritDoc}
     */
    public void providesNeg(Set<String> result) {
        assert subformula instanceof IProposition;
        //add the value that providesPos would return on the nonnegated proposition
        ((IProposition)subformula).providesPos(result); 
    }

    /**
     * {@inheritDoc}
     */
    public IAFAState reconstruct(IFormula t) {
        return factory.Not(t);
    }

    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return "!";
    }
}
