package rwth.i2.ltlrv.afastate.impl;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Map.Entry;

import rwth.i2.ltlrv.afastate.base.NullaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IIfClosure;
import rwth.i2.ltlrv.afastate.interfaze.IProposition;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesHashMap;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.util.SetUtils;

/**
 * Proposition - Implements a proposition. A proposition may hold a list of bindings,
 * which map formal parameters represented by Strings to actual objects.
 * Two propositions are equal if their label is equal and their bindings are equal as well.
 * This means, that each formal must be bound to <i>the same</i> object, not just to <i>equal</i>
 * objects. References to objects have to be provided by maps emplyoing weak references. 
 * 
 * @author Eric Bodden
 * @see java.lang.ref.WeakReference
 */
public class Proposition extends NullaryAFAState implements IProposition {

    /**
     * An appropriate label for this proposition.
     * This must be different for propositions with different semantics.
     */
    protected final String label;
    
    /**
     * Bindings of formal parameters for this proposition.
     * Must provide support for weak keys. 
     */
    protected WeakValuesMap<String, Object> bindings;
    
    /**
     * Copy of size of the bindings map. The size of the map
     * can implicitly change when objects are garbage collected.
     * This copy is used in order to detect such a change. 
     */
    protected final int bindingsSize;

    /** 
     * List of if-closures that have to be satisfied in order to match.
     * @see IIfClosure      
     */
    //TODO use Set here?
    protected final Collection<IIfClosure> ifClosures;

    //TODO still used? (depends on unboundVariablesAt...)
    protected final Set<String> varsInIfClosure;
    
    /** 
     * List of provided variables.
     * This has to be stored at creation time, since bindings change over time.
     */
    protected final String[] provides;

    /**
     * Constructs a new proposition with the given label and free bindings
     * for the given formals.
     * @param label an appropriate label for this proposition
     * @param formals list of formals that may be bound during evaluation
     * @param ifClosures if-closures which are to be evaluated at runtime
     * @see IIfClosure
     */
    public Proposition(String label, String[] formals, IIfClosure[] ifClosures) {
        this.bindings = new WeakValuesHashMap<String, Object>();
        for (String formal : formals) {
            this.bindings.put(formal, IProposition.UNBOUND);
        }
        this.bindingsSize = formals.length; 
        this.label = label;
        this.ifClosures = new ArrayList<IIfClosure>();
        this.ifClosures.addAll(Arrays.asList(ifClosures));
        this.varsInIfClosure = variableNamesInIfClosures();
        for(String formal: this.varsInIfClosure) {
            this.bindings.put(formal, IProposition.UNBOUND);
        }
        this.bindings.lock();
        this.provides = formals.clone();
    }
    
    /**
     * Constructs a new proposition with the given label and bindings of formals
     * to actual objects.<br>
     * This is deprecated. Use {@link #Proposition(String, String[], IIfClosure[])} in combination
     * with {@link #specializeBindings(WeakValuesMap)} instead.
     * @param label an appropriate label for this proposition
     * @param bindings a map mapping from formals represented as String to actual
     * objects
     */
    @Deprecated
    public Proposition(String label, WeakValuesMap<String,Object> bindings) {
        this(label, bindings, bindings.keySet().toArray(new String[0]), Collections.<IIfClosure>emptyList());
    }
    
    /**
     * Constructs a new proposition with the given label and bindings of formals
     * to actual objects.
     * This is used by {@link #specializeBindings(WeakValuesMap)}
     * @param label an appropriate label for this proposition
     * @param bindings a map mapping from formals represented as String to actual
     * @param provides old <i>provides</i> set. This is needed, cause bindings may have changed.
     * @param ifClosures if-closures which are to be evaluated at runtime
     */
    private Proposition(String label, WeakValuesMap<String, Object> bindings, String [] provides, Collection<IIfClosure> ifClosures) {
        this.bindings = bindings;
        this.bindingsSize = bindings.size(); 
        this.label = label;
        this.ifClosures = ifClosures;
        this.varsInIfClosure = variableNamesInIfClosures();
        for(String formal: this.varsInIfClosure) {
            if(!this.bindings.containsKey(formal)) {
                this.bindings.put(formal, IProposition.UNBOUND);
            }
        }
        this.bindings.lock();
        this.provides = provides;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        //if bound objects were lost, return 'FF', because
        //this proposition then behaves like FF
        if(wasGarbageCollected()) {
            return "'" + factory.FF().toString() + "'";
        }

        String s = label;
        if(bindings.keySet().size()>0) {
            s += "(" + bindings + ")"; 
        }
        if(ifClosures!=null && ifClosures.size()>0) {
            s += "(" + ifClosures + ")"; 
        } 
        return s;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
        //propositional states are nonfinal states in an AFA since the
        //proposition is still required to occur
        return false;
    }

    /**
     * Returns the set of variables that are contained in if-closures
     * (including those that are bound by this pointcut).
     * @return the set of variables that are contained in if-closures
     */
    private Set<String> variableNamesInIfClosures() {
        Set<String> result = new HashSet<String>();
        for (IIfClosure ifClosure : ifClosures) {
            result.addAll(Arrays.asList(ifClosure.variableNames()));
        }
        return result;
    }    

    /**
     * {@inheritDoc}
     */
    public boolean matches(IProposition specializedProposition) {
        //if one of the bound objects was garbage collected, this proposition is never going to
        //match again
        if(wasGarbageCollected()) {
           return false;
        }
        Proposition p = (Proposition) specializedProposition;
        //check for equal labels
        if(!this.label.equals(p.label)) {
            return false;
        }
        //check for equal number of parameters
        if(this.bindings.keySet().size()!=p.bindings.keySet().size()) {
            return false;
        }
        //check that each bound entry is identical if both are bound
        for (Entry<String,Object> entry : bindings.entrySet()) {
            Object bindingInP = p.bindings.get(entry.getKey());
            if(bindingInP!=null) {
                if(entry.getValue()!=IProposition.UNBOUND && bindingInP!=IProposition.UNBOUND) {
                    if(entry.getValue() != bindingInP) {
                        return false;
                    }
                }                
            } else {
                //something which is bound here, is missing in p
                //so this is more special than p w.r.t. this binding
                //and thus we don't match
                return false; 
            }             
        }     
        //else, we have a match
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    //we are equal if label, bindings, if closures and "provides" set are equal
    public boolean equals(Object oth) {
        if (this == oth) {
            return true;
        }
    
        if(!(oth instanceof Proposition)) {
            return false;
        }
        Proposition other = (Proposition) oth;
        if (this.bindings == null) {
            if (other.bindings != null) {
                return false;
            }
        } else {
            if (!this.bindings.equals(other.bindings)) {
                return false;
            }
        }
        if (this.label == null) {
            if (other.label != null) {
                return false;
            }
        } else {
            if (!this.label.equals(other.label)) {
                return false;
            }
        }
        if (this.provides == null) {
            if (other.provides != null) {
                return false;
            }
        } else {
            if (!Arrays.equals(provides, other.provides)) {
                return false;
            }
        }
        if (this.ifClosures == null) {
            if (other.ifClosures != null) {
                return false;
            }
        } else {
            if (!this.ifClosures.equals(other.ifClosures)) {
                return false;
            }
        }
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    //calculate over label, bindings and "provides" set 
    public int hashCode() {
        final int PRIME = 1000003;
        int result = PRIME;
        if (bindings != null) {
            result = PRIME * result + bindings.hashCode();
        }
        if (label != null) {
            result = PRIME * result + label.hashCode();
        }
        if (provides != null) {
            result = PRIME * result + Arrays.hashCode(provides);
        }
    
        return result;
    }

    /**
     * {@inheritDoc}
     */
    public WeakValuesMap<String, Object> getBindings() {
        WeakValuesMap<String, Object> copy = new WeakValuesHashMap<String, Object>();
        copy.putAll(bindings);
        return copy;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void doTransition(
            //context is unused here, since it is only used to filter bindings higher up
            @SuppressWarnings("unused") Set<String> context,
            PropositionSet currentProps,
            WeakValuesMap<String, Object> currentBinding,
            Set<Set<IAFAState>> result) {       
        
        //if one of the bound objects was garbage collected, this proposition is never going to
        //match again
        if(wasGarbageCollected()) {
           return;
        }
        
        //"return true" if this proposition matches one of the given proposition under the given bindings  
        Proposition specialized = (Proposition)this.specializeBindings(currentBinding);
        if(specialized.allIfClosuresSatisfied()) {
            //if all constraint are satisfied
            if(currentProps.containsMatchFor(specialized)) {
                //if true, add the empty set representing TT
                result.add(SetUtils.clauseTT());
            }
        } 
        
        //when this assertion fails, this means that something
        //is wrong with the backend: a proposition which is not fully bound may exist,
        //however it may never match; the formula should have been detected as invalid before
        assert (!currentProps.containsMatchFor(specialized) || specialized.fullyBound());

    }

    /**
     * @return true if a binding was garbage collected
     */
    private boolean wasGarbageCollected() {
        return this.bindings.size()<this.bindingsSize;
    }

    /**
     * Returns <code>true</code> if all if closures are currently satisfied.
     * This implies that the proposition is fully bound.
     * @return <code>true</code> if all if closures are currently satisfied
     */
    private boolean allIfClosuresSatisfied() {
        //easy case
        if(ifClosures.size()==0) {
            return true;
        }
        //if closures should only evaluate to true, when this is fully bound 
        if(!fullyBound()) {
            return false;
        }
        //check them all
        if(ifClosures!=null) {
            for (IIfClosure ifClosure : ifClosures) {
                if(!ifClosure.satisfiedUnderBindings(bindings)) {
                    return false;
                }
            }
        }
        //all went fine
        return true;
    }
    
    /**
     * Returns <code>true</code> if this proposition is fully bound.
     * @return <code>true</code> if for all <i>b</i> in {@link #bindings},
     * <i>b!=</i>{@link IProposition#UNBOUND}
     */
    private boolean fullyBound() {
        for(Object value: bindings.values()) {
            if(value==IProposition.UNBOUND) {
                return false;
            }
        }
        return true;
    }
    
    /**
     * {@inheritDoc}
     */    
    //TODO still used?
    @Override
    public void unboundVariablesAtCurrentJoinpoint(VariableKind variableKind, Set<String> result) {     
        Set<String> temp = new HashSet<String>();
        //add all unbound variables 
        for (Entry<String,Object> entry : bindings.entrySet()) {
            if(entry.getValue()==UNBOUND) {
                temp.add(entry.getKey());
            }
        }
        if(variableKind == VariableKind.IF_CLOSURES_ONLY) {
            //remove variables not contained in if-closures
            for (Iterator<String> iter = temp.iterator(); iter.hasNext();) {
                String varName = (String) iter.next();
                if(!varsInIfClosure.contains(varName)) {
                    iter.remove();
                }
            }
        }
        result.addAll(temp);
    }

    /** 
     * @inheritDoc
     */
    @Override
    public IProposition specializeBindings(WeakValuesMap<String, Object> specializedBindings) {
        //map of new bindings 
        WeakValuesMap<String,Object> newBindings = new WeakValuesHashMap<String,Object>();
        //for all formals in this proposition
        for (String formal : bindings.keySet()) {
            //if specializedBindings has a binding for this formal
            if(specializedBindings.containsKey(formal)
                    && specializedBindings.get(formal) != IProposition.UNBOUND) {
                //if the formal is still unbound in "this"
                if(bindings.get(formal)==IProposition.UNBOUND) {
                  //put the new actual binding into the map 
                  newBindings.put(formal,specializedBindings.get(formal));
                } else {
                    //the formal is already bound; this should not happen, should it?
                    newBindings.put(formal,bindings.get(formal));
                    //throw new RuntimeException("Ambiguous binding!");
                }
            }
            //else, maintain the value 
            if(!newBindings.containsKey(formal)) {
                //maintain the old value
                newBindings.put(formal,bindings.get(formal));
            }
        }
        //return a proposition with the new bindings
        return new Proposition(label, newBindings, provides, ifClosures);
    }

    /** 
     * @inheritDoc
     */
    @Override
    public void providesPos(Set<String> result) {
        assert provides != null;
        result.addAll(Arrays.asList(provides));
    }

    /** 
     * @inheritDoc
     */
    @Override
    public void providesNeg(@SuppressWarnings("unused") Set<String> result) {
        //a nonnegated proposition should not contribute anything here;
        //for negated propositions, "Not" will reroute the call to
        //providesPos
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void requires(Set<String> result) {
        Set<String> temp = new HashSet<String>();
        //we require all variables in if-closures...
        temp.addAll(varsInIfClosure);
        //... but not actually those that we provide ourselves
        temp.removeAll(Arrays.asList(provides));
        //return
        result.addAll(temp);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void validate(Set<String> context) throws ValidationException {
        //update the context according to the rules
        Set<String> localContext = updateContext(context);
        //get the required variables
        Set<String> req = new HashSet<String>();       
        requires(req);
        //check
        for (String varName : req) {
            if(!localContext.contains(varName)) {
                 throw new ValidationException(varName,this);
            }
        }
    }

    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return label;
    }
}
