package rwth.i2.ltlrv.afastate.impl;

import java.util.Set;

import rwth.i2.ltlrv.afastate.base.AbstractAFAState;
import rwth.i2.ltlrv.afastate.base.BinaryAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IFF;
import rwth.i2.ltlrv.afastate.interfaze.IRelease;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IFormula;

/**
 * Release - Implements the <i>Release</i> operator in LTL.
 * <i>Release</i> and <i>Until</i> subformulae are the essence of
 * an AFA representation of a LTL formula. They represent
 * the actual states of an AFA.s 
 * 
 * @author Eric Bodden
 * @see Until
 */
public class Release extends BinaryAFAState implements IRelease {

    /**
     * Constructs a new formula with the given subformulae.
     * @param subFormula1 left subformula
     * @param subFormula2 right subformula
     */
    public Release(IFormula subFormula1, IFormula subFormula2) {
        super(subFormula1,subFormula2);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public IAFAState negationNormalForm() {
    	//simply recurse down
    	IRelease result =  (IRelease)super.negationNormalForm();

        //remove duplicate operators
    	if (result.getSubformula2() instanceof IRelease) {
    		// phi R (phi R psi) == phi R psi
			IRelease s2 = (IRelease) result.getSubformula2();
			if(s2.getSubformula1().equals(subformula1)) {
		        return factory.Release(
		                s2.getSubformula1().negationNormalForm(),
		                s2.getSubformula2().negationNormalForm()
		                );
			}
		}
    	
    	return result;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isFinalStateInAFA() {
		//release states are always final in an AFA
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void doTransition(Set<String> context, PropositionSet currentProps, WeakValuesMap<String, Object> currentBinding, Set<Set<IAFAState>> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 
        assert (this.subformula1 instanceof AbstractAFAState) && (this.subformula2 instanceof AbstractAFAState);
        
        //delta(phi R psi) = (delta(psi) &&  delta(phi)) || [delta(psi) && X(phi R psi)]
        factory.Or(
                factory.And(
                		((IAFAState)subformula2),
                        ((IAFAState)subformula1)
                ),
                factory.And(
                		((IAFAState)subformula2),
                        factory.X(this)
                )        
        ).transition(context,currentProps,currentBinding, result);
	}
    
	/**
	 * {@inheritDoc}
	 */
	public void providesPos(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 

        //providesPos( phi R psi ) =  providesPos( (psi /\ phi) \/ (psi /\ X(phi R psi)) )
        //                         =  providesPos( psi )
        ((IAFAState)subformula2).providesPos(result);
	}
    
    /**
     * {@inheritDoc}
     */
    public void providesNeg(Set<String> result) {
        //only subformulae implementing IAFAState allowed at this stage
        assert (this.subformula1 instanceof IAFAState) && (this.subformula2 instanceof IAFAState); 
        
        //providesNeg( phi R psi ) =  providesNeg( psi ) intersect providesNeg( phi )
        //                         =  providesNeg( phi /\ psi )
        factory.And(subformula2,subformula1).providesNeg(result);
    }
   
    /**
     * {@inheritDoc}
     */
    public IAFAState reconstruct(IFormula t1, IFormula t2) {
        return factory.Release(t1, t2);
    }

    /**
     * {@inheritDoc}
     */
    public String symbol() {
        return "R";
    } 
    
    /**
     * {@inheritDoc}
     */
    @Override    
    public String toString() {      
        if(this.subformula1 instanceof IFF) {
            //render (FF R psi) as G(psi)
            return factory.G(this.subformula2).toString();
        } else {
            return super.toString();
        }
    }
    
}
