package rwth.i2.ltlrv.data;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import rwth.i2.ltlrv.afastate.interfaze.IProposition;

/**
 * PropositionSet - A set of proposition with some additional functionality.
 * 
 * @author Eric Bodden
 */
@SuppressWarnings("serial")
public class PropositionSet extends HashSet<IProposition> {
    
    /**
     * Constructs an empty set.
     */
    public PropositionSet() {
        super();
    }
    
    /**
     * Constructs a new set based on the collection of propositions
     * given as input.
     * @param propositions propositions the set should hold
     */
    public PropositionSet(Collection<IProposition> propositions) {
        super();
        addAll(propositions);
    }    
    
    /**
     * Returns the unique proposition equal to <i>proposition</i>
     * which is contained in this set or <code>null</code> if
     * there is not such proposition.
     * @param proposition search key
     * @return proposition equal to <i>proposition</i> or <code>null</code>
     *  if there is none
     */
    public IProposition getEqual(IProposition proposition) {
        for (Iterator<IProposition> iter = super.iterator(); iter.hasNext();) {
            IProposition p = iter.next();
            if(proposition.equals(p)) {
                return p;
            }
        }
        return null;
    }
        
    /**
     * Returns <code>true</code> if this set contains a proposition which
     * is matched by the given proposition.
     * @param prop the proposition which maybe matches one of the propositions
     * of this set
     * @return <code>true</code> if the proposition does match one of the contained propositions
     * @see IProposition#matches(IProposition)
     */
    public boolean containsMatchFor(IProposition prop) {
        for(IProposition p : this) {
            if(prop.matches(p)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Returns an iterable over all valid combinations of bindings. That is the set product of all sets of bindings.
     * for example, be given the bindings <i>a=a1</i>, <i>a=a2</i> and <i>b=b1</i>, this will result in mappings
     * <i>a=a1, b=b1</i> and <i>a=a2, b=b1</i>. 
     * @return an iterable over all valid combinations of bindings
     */
    public Iterable<WeakValuesMap<String, Object>> validCombinationsOfBindings() {
        WeakValuesMap<String, Set<Object>> containedBindings = containedBindings();
        return validCombinations(containedBindings);
    }
    
    /**
     * Returns the set of contained bindings in the form <i>a={a2,a2},b={b2},...</i> for bindings
     * <i>a=a1</i>, <i>a=a2</i> and <i>b=b1</i>.
     * @return the set of contained bindings
     */
    private WeakValuesMap<String, Set<Object>> containedBindings() {
        WeakValuesMap<String, Set<Object>> allBindings = new WeakValuesHashMap<String, Set<Object>>();
        //over all contained propositions
        for (IProposition prop : this) {
            WeakValuesMap<String, Object> bindings = prop.getBindings();
            //over all bindings for this proposition
            //fold those bindings
            for (Entry<String,Object> entry : bindings.entrySet()) {
            	String key = entry.getKey();
            	Object val = entry.getValue();
            	//if value is not UNBOUND
                if(!(val==IProposition.UNBOUND)) {
                	Set<Object> bindingsForKey = allBindings.get(key);
                	//create new set if necessary
                    if(bindingsForKey==null) {
                    	bindingsForKey = new HashSet<Object>();
                        allBindings.put(key, bindingsForKey);
                    }
                    //add to the set
                    bindingsForKey.add(val);
                }
            }
        }
        return allBindings;        
    }
    
    /**
     * Returns an iterable over all valid combinations of a set of bindings. That is the set product of all sets of bindings.
     * for example, be given the bindings <i>a=a1</i>, <i>a=a2</i> and <i>b=b1</i>, this will result in mappings
     * <i>a=a1, b=b1</i> and <i>a=a2, b=b1</i>. 
     * @param bindings the set of bindings in the form <i>a={a2,a2},b={b2},...</i>
     * @return an iterable over all valid combinations of bindings
     */
    private Collection<WeakValuesMap<String, Object>> validCombinations(
            WeakValuesMap<String, Set<Object>> bindings) {
        //construct result set
        Collection<WeakValuesMap<String, Object>> result = new ArrayList<WeakValuesMap<String, Object>>();
        //accumulate results
        validCombinations(bindings, new HashMap<String, Object>(), result);
        return result;
    }

    /**
     * Returns an iterable over all valid combinations of a set of bindings. That is the set product of all sets of bindings.
     * for example, be given the bindings <i>a=a1</i>, <i>a=a2</i> and <i>b=b1</i>, this will result in mappings
     * <i>a=a1, b=b1</i> and <i>a=a2, b=b1</i>. 
     * @param bindings the set of bindings in the form <i>a={a2,a2},b={b2},...</i>
     * @param accummulator accumulator, for temporary use only, must be empty at the beginning 
     * @param result the result is placed here
     */
    private void validCombinations(WeakValuesMap<String, Set<Object>> bindings,
            Map<String, Object> accummulator,
            Collection<WeakValuesMap<String, Object>> result) {
        
        //splice away the first mapping
        Set<Entry<String, Set<Object>>> entries = bindings.entrySet();
        Iterator<Entry<String, Set<Object>>> iter = entries.iterator();
        if (iter.hasNext()) {
            //if we have further mappings
            Entry<String, Set<Object>> entry = iter.next();
            bindings.remove(entry.getKey());
            for (Object obj : entry.getValue()) {
                accummulator.put(entry.getKey(), obj);
                //recurse
                validCombinations(bindings, accummulator, result);
                accummulator.remove(entry.getKey());
            }
            bindings.put(entry.getKey(), entry.getValue());
        } else {
            //simply return the current state
            WeakValuesMap<String, Object> newMap = new WeakValuesHashMap<String, Object>();
            newMap.putAll(accummulator);
            result.add(newMap);
        }
    }    


    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("\n[\n");

            Iterator<IProposition> i = iterator();
            boolean hasNext = i.hasNext();
            while (hasNext) {
                IProposition o = i.next();
                buf.append(String.valueOf(o));
                hasNext = i.hasNext();
                if (hasNext)
                    buf.append(",\n\n");
            }

        buf.append("\n]");
        return buf.toString();
    }
}
