package rwth.i2.ltlrv.management;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;
import java.util.Map.Entry;

import rwth.i2.ltlrv.afastate.interfaze.IAFAState;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState.ValidationException;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.data.WeakValuesHashMap;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.util.SetUtils;

/**
 * Configuration - Global configuration for one single formula.
 * 
 * @author Eric Bodden
 */
public class Configuration {

    /** The current global state as a disjunct of conjuncts of terms. */
    private Set<Set<IAFAState>> globalState;

    /** Maps a term to a number. For debugging purposes only. */
    private static Map<IAFAState,Integer> termToNumber = new WeakHashMap<IAFAState,Integer>();
    
    /** Last number a term was mapped to. For debugging purposes only. */
    private static int lastTermNumber = 0;        
    
    /**
     * Constructs a new configuration for an initial state.
     * 
     * @param initialState
     *            the initial state of the system
     * @throws ValidationException thrown if there are inconsistent bindings
     * 	in this formula
     */
    public Configuration(IAFAState initialState) throws ValidationException {
    	//TODO HANDLE CORRECTLY
    	try {
            initialState.validate();
        } catch (ValidationException e) {
            System.err.println(initialState);
            throw e;
        }
        globalState = SetUtils.singletonClauseSet(initialState);
    }

    /**
     * Performs a transition for the given propositions.
     * 
     * @param props
     *            a set of runtime propositions
     */
    /* package*/ void transition(PropositionSet props) {
        //TODO nondestructive update here
        globalState = transition(globalState, props);
    }

    /**
     * Returns true if this configuration is currently in a final (accepting) state.
     * @return true iff the configuration is final/accepting
     */
    public boolean isFinal() {
        return isClauseSetFinal(globalState);
    }

    /**
     * @param clauseSet the clause set to check
     * @return <code>true</code> is this clause set represents a final state
     */
    private static boolean isClauseSetFinal(Set<Set<IAFAState>> clauseSet) {
        for (Set<IAFAState> clause : clauseSet) {
            boolean clauseFinal = true; 
            for (IAFAState term : clause) {
                if(!term.isFinalStateInAFA()) {
                    clauseFinal = false;
                    break;
                }
            }
            if(clauseFinal) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Returns true if this configuration is in the global state TT.
     * @return <code>true</code> if the global state of this configuration
     *  is {{}} 
     */
    public boolean isTT() {
        return globalState.equals(SetUtils.TT());
    }
    
    /**
     * Returns true if this configuration is in the global state FF.
     * @return <code>true</code> if the global state of this configuration
     *  is {} 
     */
    public boolean isFF() {
        return globalState.equals(SetUtils.FF());
    }

    /**
     * Performs a transition from the given state (set of clauses) to the sucessor state
     * for the given set of propositions. 
     * @param globalState the state to calculate the successor for
     * @param currentProps the set of propositions to calculate the transition for
     * @return the new state (set of clauses)
     */
    private static Set<Set<IAFAState>> transition(Set<Set<IAFAState>> globalState,
            PropositionSet currentProps) {
              
        //calculate all valid bindings
        final Iterable<WeakValuesMap<String, Object>> validBindings =
            currentProps.validCombinationsOfBindings();

        //construct new result set
        Set<Set<IAFAState>> newGlobalState = SetUtils.FF();
        
        //over all clauses
        for (Set<IAFAState> setOfStates : globalState) {

            //construct new result set, initialized to the representation of "true"
            Set<Set<IAFAState>> resultForOneClause = SetUtils.TT();
            
            //over all states
            for (IAFAState state : setOfStates) {
                
                //find unbound variables at current jooinpoint
                final Set<String> unboundVariablesAtCurrentJoinpoint = new HashSet<String>();
                state.unboundVariablesAtCurrentJoinpoint(IAFAState.VariableKind.ALL, unboundVariablesAtCurrentJoinpoint);
                
                //find unbound variables in if closures at current jooinpoint
                final Set<String> unboundVariablesInIfClosuresAtCurrentJoinpoint = new HashSet<String>();
                state.unboundVariablesAtCurrentJoinpoint(IAFAState.VariableKind.IF_CLOSURES_ONLY, unboundVariablesInIfClosuresAtCurrentJoinpoint);
                
                //construct new result set, initialized to the representation of "true"
                Set<Set<IAFAState>> rawResult = SetUtils.TT();
                
                //over all valid bindings
                for (WeakValuesMap<String, Object> binding : validBindings) {
//                	boolean allBound = true;
//                	for(String varName: unboundVariablesInIfClosuresAtCurrentJoinpoint) {
//                		if(!binding.keySet().contains(varName) || binding.get(varName)==IProposition.UNBOUND) {
//                			allBound = false;
//                			break;
//                		}
//                	}
//                	if(!allBound) {
//                		continue;
//                	}
                	
                    binding = filterBindingForCurrentJoinpoint(binding,unboundVariablesAtCurrentJoinpoint);
                    
                    //start with no clauses
                    Set<Set<IAFAState>> singleBindingResult = SetUtils.emptyClauseSet();
                    //calculate the transition under the current binding and propositions
					state.transition(new HashSet<String>(), currentProps, binding, singleBindingResult);

					//TODO remove
                    //specialize bindings in newly gained states
                    //state.specializeBindings(binding, singleBindingResult);
                    //filter out states which do not satisfy all if closures 
                    //singleBindingResult = filterForValidStates(singleBindingResult);                    
                    
                    //construct new conjunct, still empty
                    Set<Set<IAFAState>> conjunct = SetUtils.emptyClauseSet();
                    //perform set multiplication
                    SetUtils.clausesetProduct(singleBindingResult, rawResult, conjunct);
                    //store result
                    rawResult = conjunct;

                    //performance improvement:
                    //if one conjunct if false, everything is false
                    if(rawResult.equals(SetUtils.FF())) {                        
                        break;
                    }
                }
                //the result may still not be the smallest possible model;
                //thus we perform a subsumption reduction
                Set<Set<IAFAState>> reducedResult = subsumptionReduction(rawResult);
                
                //System.err.println("STATE '"+numberStringOf(state)+"' PRODUCED \n"+clauseSetToString(reducedResult));
                
                //construct new conjunct, still empty
                Set<Set<IAFAState>> conjunct = SetUtils.emptyClauseSet();
                //perform set multiplication
                SetUtils.clausesetProduct(resultForOneClause, reducedResult, conjunct);

                //store result
                resultForOneClause = conjunct;

                //performance improvement:
                //if one conjunct if false, everything is false
                if(resultForOneClause.equals(SetUtils.FF())) {                        
                    break;
                }
            }
            //the new global state is a disjunct of all results of all clauses 
            newGlobalState.addAll(resultForOneClause);
        }
        return subsumptionReduction(newGlobalState);
    }

//	/**
//	 * @param singleBindingResult
//	 */
//	private static Set<Set<IAFAState>> filterForValidStates(Set<Set<IAFAState>> singleBindingResult) {
//		Set<Set<IAFAState>> result = new HashSet<Set<IAFAState>>();
//		for (Iterator<Set<IAFAState>> setIter = singleBindingResult.iterator(); setIter.hasNext();) {
//			Set<IAFAState> set = setIter.next();
//			boolean wasTT = set.equals(SetUtils.clauseTT());
//			for (Iterator<IAFAState> termIter = set.iterator(); termIter.hasNext();) {
//				IAFAState term = termIter.next();
//				if(!term.isValidState()) {
//					termIter.remove();
//				}
//			}
//			if(wasTT || set.size()>0) {
//				result.add(set);
//			}
//		}
//		return result;
//	}

    /**
     * Filters a binding for those entries that are unbound at the current joinpoint.
     * @param binding A full binding at the current joinpoint.
     * @param unboundVariablesAtCurrentJoinpoint collection of variabels that are unbound at
     *  the current joinpoint in the given formula
     * @return the filtered map of bindings
     */
    private static WeakValuesMap<String, Object> filterBindingForCurrentJoinpoint(
            WeakValuesMap<String, Object> binding, Collection<String> unboundVariablesAtCurrentJoinpoint) {
        WeakValuesMap<String, Object> newBinding = new WeakValuesHashMap<String,Object>();
        for (Entry<String,Object> entry : binding.entrySet()) {
            if(unboundVariablesAtCurrentJoinpoint.contains(entry.getKey())) {
                newBinding.put(entry.getKey(),entry.getValue());
            }
        }
        return newBinding;
    }
    
    /**
     * Removes sets which contain another set (in order to find the smallest
     * model)
     * @param set the set on which the reduction should be performed 
     * @return the redueced set
     */
    private static Set<Set<IAFAState>> subsumptionReduction(Set<Set<IAFAState>> set) {
        if (set.size() < 2) {
            return set;
        }
        Set<Set<IAFAState>> result;
        result = new HashSet<Set<IAFAState>>();
        for (Set<IAFAState> set1 : set) {
            boolean containsAnotherSet = false;
            for (Set<IAFAState> set2 : set) {
                if (set1 != set2 && set1.containsAll(set2)) {
                    containsAnotherSet = true;
                    break;
                }
            }
            if (!containsAnotherSet) {
                result.add(set1);
            }
        }
        return result;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public String toString() {
        return clauseSetToString(globalState);
    }

    /**
     * @param clauseSet set to return a string for
     * @return a string representation of the clause set
     */
    private static String clauseSetToString(Set<Set<IAFAState>> clauseSet) {
        StringBuffer stateString = new StringBuffer();
        stateString.append("\n[\n");
        for (Set<IAFAState> set : clauseSet) {
            stateString.append("    [\n");
            for (Iterator<IAFAState> iter = set.iterator(); iter.hasNext();) {
                IAFAState term = iter.next();
                stateString.append("        ");
                stateString.append(numberStringOf(term));
                stateString.append(" ");
                stateString.append(term.isFinalStateInAFA()?"[final]   ":"[nonfinal]");
                stateString.append(" ");
                stateString.append(term);
                if(iter.hasNext()) {
                    stateString.append("\n\n");
                }
            }
            stateString.append("\n    ]\n");
        }
        stateString.append("]\n");
        
        return (isClauseSetFinal(clauseSet)?"[final]":"[nonfinal]") + "  " +  stateString;
    }
    
    /**
     * Returns a (possibly new) number for a term, based on equality.
     * @param term a term
     * @return a (possibly new) unique number for this term
     */
    private static int numberOf(IAFAState term) {        
        Integer cached = termToNumber.get(term);
        if(cached!=null) {
            return cached;
        } else {
            termToNumber.put(term,lastTermNumber);
            lastTermNumber++;
            return (lastTermNumber-1);
        }    
    }
    
    /**
     * Returns a unique number string for the term, based on equality.
     * @param term a term
     * @return a unique number string for this term
     */
    private static String numberStringOf(IAFAState term) {
        String result = "#"+numberOf(term);
        for(int i=result.length(); i<4; i++) {
            result = " "+result;
        }
        return result;
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public boolean equals(Object oth) {
        if (this == oth) {
            return true;
        }
    
        if (oth == null) {
            return false;
        }
    
        if (oth.getClass() != getClass()) {
            return false;
        }
    
        Configuration other = (Configuration) oth;
        if (this.globalState == null) {
            if (other.globalState != null) {
                return false;
            }
        } else {
            if (!this.globalState.equals(other.globalState)) {
                return false;
            }
        }
    
        return true;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public int hashCode() {
        final int PRIME = 1000003;
        int result = 0;
        if (globalState != null) {
            result = PRIME * result + globalState.hashCode();
        }
    
        return result;
    }

}
