package rwth.i2.ltlrv.management;

import rwth.i2.ltlrv.afastate.interfaze.IAnd;
import rwth.i2.ltlrv.afastate.interfaze.IFF;
import rwth.i2.ltlrv.afastate.interfaze.IIfClosure;
import rwth.i2.ltlrv.afastate.interfaze.INext;
import rwth.i2.ltlrv.afastate.interfaze.INot;
import rwth.i2.ltlrv.afastate.interfaze.IOr;
import rwth.i2.ltlrv.afastate.interfaze.IProposition;
import rwth.i2.ltlrv.afastate.interfaze.IRelease;
import rwth.i2.ltlrv.afastate.interfaze.ITT;
import rwth.i2.ltlrv.afastate.interfaze.IUntil;
import rwth.i2.ltlrv.data.WeakValuesMap;
import rwth.i2.ltlrv.formula.interfaze.IEquivalent;
import rwth.i2.ltlrv.formula.interfaze.IFinally;
import rwth.i2.ltlrv.formula.interfaze.IFormula;
import rwth.i2.ltlrv.formula.interfaze.IGlobally;
import rwth.i2.ltlrv.formula.interfaze.IImplies;

/**
 * IFormulaFactory - Common interface for formula factories.
 * Used ot instantiate formulae of any kind.
 *
 * @author Eric Bodden
 */
public interface IFormulaFactory {
    
    /**
     * Generates an conjunct.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 /\ formula2</code>
     */
    public IAnd And(IFormula formula1, IFormula formula2);

    /**
     * Generates an disjunct.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 \/ formula2</code>
     */
    public IOr Or(IFormula formula1, IFormula formula2);

    /**
     * Generates an implication.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 -> formula2</code>
     */
    public IImplies Impl(IFormula formula1, IFormula formula2);
    
    /**
     * Generates an equivalence relation.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 <-> formula2</code>
     */
    public IEquivalent Eq(IFormula formula1, IFormula formula2);
        
    /**
     * Generates a Release formula.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 R formula2</code>
     */
    public IRelease Release(IFormula formula1, IFormula formula2);
    
    /**
     * Generates an Until formula.
     * @param formula1 left subformula
     * @param formula2 right subformula
     * @return <code>formula1 U formula2</code>
     */
    public IUntil Until(IFormula formula1, IFormula formula2);
    
    /**
     * Generates a Finally formula.
     * @param formula the subformula
     * @return <code>F subformula</code>
     */
    public IFinally F(IFormula formula);
    
    /**
     * Generates a Globally formula.
     * @param formula the subformula
     * @return <code>G subformula</code>
     */
    public IGlobally G(IFormula formula);

    /**
     * Generates a Next formula.
     * @param formula the subformula
     * @return <code>X subformula</code>
     */
    public INext X(IFormula formula);

    /**
     * Generates a negation.
     * @param formula the subformula
     * @return <code>! subformula</code>
     */
    public INot Not(IFormula formula);
    
    /**
     * Returns the unique positive sink state.
     * @return <code>TT</code>
     */
    public ITT TT();
    
    /**
     * Returns the negative positive sink state.
     * @return <code>TT</code>
     */
    public IFF FF();
    
    /**
     * Returns a proposition with the given parameters.
     * @param label a label; this should be distinct for semantically distinct propositions
     * @param boundFormals the formals which will be bound by this pointcut at runtime
     * @return an appropriate proposition with <code>boundFormals</code> initialized to
     * {@link IProposition#UNBOUND}
     */
    public IProposition Proposition(String label, String... boundFormals);

    /**
     * Returns a proposition with the given parameters.
     * @param label a label; this should be distinct for semantically distinct propositions
     * @param boundFormals the formals which will be bound by this pointcut at runtime
     * @param ifClosures the if-closures which this pointcut has to satisfy
     * (@see IIfClosure)
     * @return an appropriate proposition with <code>boundFormals</code> initialized to
     * {@link IProposition#UNBOUND}
     */
    public IProposition Proposition(String label, String[] boundFormals, IIfClosure[] ifClosures);

    /**
     * This is only to be used by the test harness!
     * @param label a label; this should be distinct for semantically distinct propositions
     * @param bindings a map of current bindings
     * @return an appropriate proposition with the given bindings 
     */
    @Deprecated
    public IProposition Proposition(String label, WeakValuesMap<String,Object> bindings);
    

}
