package rwth.i2.ltlrv.management;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import rwth.i2.ltlrv.afastate.interfaze.IIfClosure;
import rwth.i2.ltlrv.afastate.interfaze.IAFAState.ValidationException;
import rwth.i2.ltlrv.data.PropositionSet;
import rwth.i2.ltlrv.formula.interfaze.IFormula;
import rwth.i2.ltlrv.management.listeners.ConsoleListener;

/**
 * VerificationRuntime - Runtime manager holding formula configurations and notifying
 * listeners about the ongoing evaluations.
 * 
 * @author Eric Bodden
 */
public class VerificationRuntime {
	
    /**
     * Listener - Interface for listers that wish to be notified about evaluation progress.
     * 
     * @author Eric Bodden
     */
    public interface Listener {
        
        /**
         * This is called whenever a new formula is registered.
         * This can happen basically any time but only once for each formula,
         * since a formula is not registered before it is actually needed (first joinpoint).
         * @param formulaId the ID of the formula
         * @param associatedThread the Thread which registered this formula 
         * @param initialConfig the initial configuration
         */
        public void notifyRegistered(String formulaId, Thread associatedThread, Configuration initialConfig);
        
        /**
         * This is called whenever a new formula is updated.
         * It may be the case that the new configuration is equal to the last one.
         * No update will be triggered after a formula evaluated to TT or FF.
         * @param formulaId the ID of the formula
         * @param associatedThread the Thread which registered this formula 
         * @param newConfig the new configuration
         */
        public void notifyUpdate(String formulaId, Thread associatedThread, Configuration newConfig);

		/**
		 * This is called when the application is going down, so after the last <code>main</code>
		 * method has been executed. For each formula the last configuration is passed.
         * Formulas which evaluated to TT/FF before, will not be passed.
         * It is up to the client to keep track of such formulae.
         * <b>IMPORTANT:</b> This will be run from inside a shutdown hook. So be sure to invoke
         * nothing which tries to modify shutdown hooks at this time. In particular, some AWT
         * classes like {@link java.awt.Toolkit} may do so.  
		 * @param formulaId
		 * @param config
		 */
		public void notifyTearDown(String formulaId, Configuration config);

        /**
         * Notifies about a user-caused exception that happened during the evaluation of an if-pointcut.
         * The formula will be removed from further evaluation. So no update for this formula will occur.
         * @param formulaId id of the formula which caused this exception 
         * @param associatedThread Thread that triggered the evaluation
         * @param ifExpression expression that caused the exception
         * @param exception exception that occured
         * @param config last configuration before this exception occured
         */
        public void notifyOnUserCauseException(String formulaId, Thread associatedThread, String ifExpression, Throwable exception, Configuration config);
        
    }
    
    /** 
     * Environment variable pointing to listener classes (semicolon separated):
     * LTLRV_LISTENERS=package1.Listener1;package1.Listener2 
     */
    private static final String LTL_RV_LISTENERS = "LTLRV_LISTENERS";
	
	/** The default listeners for the case that the environment variable is not set. */
	private static final Listener[] DEFAULT_LISTENERS = { new ConsoleListener() };

	/** Singleton instance of this class. */
	private static VerificationRuntime singletonInstance = null;
	
	/** Mapping from formula ID to the configuration fot this formula. */
	private Map<String,Configuration> configurationForFormulaId;
	
	/** Similar to {@link #configurationForFormulaId} but keeps copies of the initial configurations */
	private Map<String,Configuration> originalConfigurationForFormulaId;

	/** All currently registered listeners. */
	private List<Listener> listeners;

	/** If rematching is on, falsified/satisfied formulae are not removed but rather reinstantiated. */
	private final static boolean REMATCHING = System.getProperty("JLO_REMATCHING")!=null;

	/**
	 * Default constructor. Called by getInstance() only.
     * @see #getInstance() 
	 */
	private VerificationRuntime() {
		this.configurationForFormulaId = new HashMap<String,Configuration>();
		if(REMATCHING) {
			this.originalConfigurationForFormulaId = new HashMap<String,Configuration>();
		}
		this.listeners = new ArrayList<Listener>();
        this.initializeListeners();
	}
	
	/**
	 * Installs listeners from the environment variable or installs the default listeners if
     * the variable is not set.
	 */
	private void initializeListeners() {
		String listenerClassesString = System.getProperty(LTL_RV_LISTENERS);
		if(listenerClassesString!=null) {
			//dynamically instantiate listener classes
			String[] listenerClasses = listenerClassesString.split(";");
			for(String listenerClass: listenerClasses) {
				try {
					Class cl = this.getClass().getClassLoader().loadClass(listenerClass);
					try {
						@SuppressWarnings("unchecked")
                        Object listenerObject = cl.newInstance();
						if(Listener.class.isAssignableFrom(listenerObject.getClass())) {
							Listener listener = (Listener) listenerObject;
							//register
							this.listeners.add(listener);
						} else {
							System.err.println("Error: Listener class " + listenerClass + " has invalid type.");
						}
					} catch (InstantiationException e1) {
						System.err.println("Error: Listener class " + listenerClass + " could not be instantiated.");
						System.err.println(e1.getMessage());
					} catch (IllegalAccessException e1) {
						System.err.println("Error: Listener class " + listenerClass + " could not be instantiated.");
						System.err.println(e1.getMessage());
					}
				} catch (ClassNotFoundException e) {
					System.err.println("Error: Listener class " + listenerClass + " not found.");
				}
				
			}			
		} else {
			//instantiate default listeners
			for(Listener listener: DEFAULT_LISTENERS) {
				this.listeners.add(listener);
			}
		}
	}

	/**
     * Registers a new formula.
	 * @param formulaId Unique ID for this formula.
	 * @param formula The new formula to install.
	 */
	public synchronized void registerFormula(String formulaId, IFormula formula) {
	    assert !this.configurationForFormulaId.keySet().contains(formula);
        Configuration config;
		try {
			config = new Configuration(formula.negationNormalForm());
            Debug.logConfig(formulaId,config);
		} catch (ValidationException e) {
			//FIXME check statically
			throw new RuntimeException(e);
		}
        if( !( config.isTT()||config.isFF() ) ) {
            //if formula is nontrivial, add it
            this.configurationForFormulaId.put(formulaId, config);
            
            if(REMATCHING) {
    			Configuration copy;
				try {
					copy = new Configuration(formula.negationNormalForm());
	    			this.originalConfigurationForFormulaId.put(formulaId,copy);
				} catch (ValidationException e) {
					//cannot actually happen cause we have validated before
					throw new RuntimeException("Internal error.");
				}
            }
        }
        this.notifyListeners(false, formulaId, Thread.currentThread(), config);
	}
    
    /**
     * Requests that the formula with the given ID shall be updated.
     * @param formulaId Unique ID for the formula to be updated.
     * @param currentPropositions the set of propositions under which the transition shall be taken.
     */
    public synchronized void updateFormula(String formulaId, PropositionSet currentPropositions) {
        assert currentPropositions.size()>0;

        Debug.logPropositions(formulaId,currentPropositions);
       
        Configuration config = configurationForFormulaId.get(formulaId);
        if(config!=null) {
            try {
            	//take transition, user-caused exception may occur here
                config.transition(currentPropositions);
                Debug.logConfig(formulaId, config);
                //if everything went fine...
                this.notifyListeners(true, formulaId, Thread.currentThread(), config);
                if(config.isTT()||config.isFF()) {
                	if(REMATCHING) {
                		//if rematching is enabled, do not drop the formula but rather reinstantiate it
                		configurationForFormulaId.put(formulaId, originalConfigurationForFormulaId.get(formulaId));
                	} else {
                        //remove configuration of formula is satisfied/falsified
                        configurationForFormulaId.remove(formulaId);                		
                	}
                }
            } catch( IIfClosure.UserCausedException ucex ) {
                //if an exception occured during evaluation of an if-pointcut,
                //notify all listeners about this
                notifyOnUserCausedException(formulaId, Thread.currentThread(), ucex.getIfExpression(), ucex.getCause(), config);
                //remove formula
                configurationForFormulaId.remove(formulaId);
            }
        }
    }
	
    /**
     * Notifies that the system shuts down.
     */
    public void tearDown() {
        for(Listener l: this.listeners) {
        	for (Entry<String,Configuration> entry : configurationForFormulaId.entrySet()) {
            	final String formulaId = entry.getKey();
            	final Configuration config = entry.getValue();
                l.notifyTearDown(formulaId,config);
            }
		}
    }
    
    /**
     * Notifies all listeners about a user-caused exception that happened during the evaluation of an if-pointcut.
     * @param formulaId id of the formula which caused this exception 
     * @param thread Thread that triggered the evaluation
     * @param ifExpression expression that caused the exception
     * @param exception exception that occured
     * @param config last configuration before this exception occured
     */
    protected void notifyOnUserCausedException(String formulaId, Thread thread, String ifExpression, Throwable exception, Configuration config) {
        for (Listener listener : listeners) {
            listener.notifyOnUserCauseException(formulaId, thread, ifExpression,exception,config);
        }
    }

    /**
     * Notifies all listeners about a formula being updated/registered.
	 * @param update if <code>true</code>, this calls <code>notifyUpdate</code>, else <code>notifyRegistered</code>
	 * @param formulaId Unique ID for the formula to be notified about.
	 * @param associatedThread Thread who updated/installed the formula.
	 * @param newConfig The new/initial configuration.
     * @see VerificationRuntime.Listener#notifyUpdate(String, Thread, Configuration)
     * @see VerificationRuntime.Listener#notifyRegistered(String, Thread, Configuration)
	 */
	private void notifyListeners(boolean update, String formulaId, Thread associatedThread, Configuration newConfig) {
        if(update) {
            for(Listener l: this.listeners) {
                l.notifyUpdate(formulaId,associatedThread,newConfig);
            }
        } else {
            for(Listener l: this.listeners) {
                l.notifyRegistered(formulaId,associatedThread,newConfig);
            }
        }
	}
	
	/**
     * Returns the singleton instance.
	 * @return the singleton instance
	 */
	public static VerificationRuntime getInstance() {
		if(singletonInstance == null) {
			singletonInstance = new VerificationRuntime();
		}
		return singletonInstance;
	}
    
    /**
     * Returns the term factory.
     * @return The term factory.
     */
    public IFormulaFactory getFactory() {
        //TODO make generic
        return FORMULA_FACTORY;
    }
    
    private final IFormulaFactory FORMULA_FACTORY = new FormulaFactory();  

}
