package rwth.i2.ltlrv.util;

import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import rwth.i2.ltlrv.afastate.interfaze.IAFAState;

/**
 * SetUtils - Some static convenience methods for set operations.
 * 
 * @author Eric Bodden
 */
public class SetUtils {
    
    /**
     * Constructs an empty clause set, which represents <code>TT</code>
     * (<code>true</code>). 
     * @return {{}}
     */
    public static Set<Set<IAFAState>> TT() {
        return singletonClauseSet();
    }
    
    /**
     * Constructs an empty clause which represents <code>TT</code>
     * (<code>true</code>). 
     * @return {}
     */
    public static Set<IAFAState> clauseTT() {
        return termClause();
    }

    /**
     * Constructs a term clause/set for the given elements. 
     * @param elements some terms
     * @return a clause holding those terms
     */
    public static Set<IAFAState> termClause(IAFAState... elements) {
        Set<IAFAState> set = new HashSet<IAFAState>();
        set.addAll(Arrays.asList(elements));
        return set;
    }
    
    /**
     * Constructs a set holding one single clause holding the given terms.
     * @param elements some terms
     * @return a set holding one clause holding the given terms
     */
    public static Set<Set<IAFAState>> singletonClauseSet(IAFAState... elements) {
        Set<Set<IAFAState>> set = new HashSet<Set<IAFAState>>();
        set.add(termClause(elements));
        return set;
    }
    
    /**
     * Constructs an empty clause set.
     * @return an empty set
     */
    public static Set<Set<IAFAState>> emptyClauseSet() {
        return new HashSet<Set<IAFAState>>();
    }

    /**
     * Constructs an empty clause set, representing FF.
     * @return an empty set
     */
    public static Set<Set<IAFAState>> FF() {
        return emptyClauseSet();
    }

    /**
     * Implements the (cartesian) product of two clause sets. If <i>set1={{a,b}}</i> and <i>set2={{c},{d}}</i>
     * the result is <i>{{a,c},{a,d},{b,c},{b,d}}</i>. In logic terms this is the law of <i>de Morgan</i>. 
     * @param <T> type parameter (optional, used to circumvent type inference) 
     * @param set1 first factor
     * @param set2 second factor 
     * @param result product of the two sets
     */
    public static <T> void clausesetProduct(Set<Set<T>> set1, Set<Set<T>> set2, Set<Set<T>> result) {
        for (Set<T> clause1 : set1) {
            for (Set<T> clause2 : set2) {
                Set<T> join = new HashSet<T>();
                join.addAll(clause1);
                join.addAll(clause2);
                result.add(join);
            }
        }
    }
}
